#!/usr/bin/env python
#
# PerfEval.py: hack to check RISC simulation performance results from logs
#
# author: Rainer Doemer
#
# modifications:
# 04/14/17 RD	adjusted performance expectations for RISC 0.4.0
# 04/27/17 RD	adjusted for examples/parallel, seq and ooo only
# 04/25/17 RD	initial version (test/regression)

# imports
import sys
import os
import time

# classes:
class Perf:
    "performance values of an example"
    def __init__(self,h,u,s,e,c,x):
	self.h = h
	self.u = u
	self.s = s
	self.e = e
	self.c = c
	self.x = x

# checks to be performed
Checks = [
    ("fibo", # always runs in synchronous mode due to sc_fifo
	("seq.x== 0", "ooo.x== 0",
	 "seq.c<=100","ooo.c>=450",
	 "seq.e<=6.0","ooo.e<=2.0")),
    ("fmul", # no dependencies, so ooo == npd == syn performance
	("seq.x== 0", "ooo.x== 0",
	 "seq.c<=100","ooo.c>=650",
	 "seq.e<=6.0","ooo.e<=1.0"))]

# functions
def PrintLogLine(d,e):
    print("%s: %0.2f user %0.2f system %0.2f elapsed %d%% CPU %d exit: %s" %
	(d.h, d.u, d.s, d.e, d.c, d.x, e))

#
# Action starts here!
#

# parse command line arguments

Cmd = os.path.basename(sys.argv[0])
if len(sys.argv) < 2:
    sys.stderr.write("%s: bad arguments!\n" % Cmd)
    sys.stderr.write("Usage: %s <log file>\n" % Cmd)
    sys.exit(10)

LogName = sys.argv[1]
sys.stderr.write("%s: Reading log file \"%s\"...\n" % (Cmd,LogName))

# read the log file

try:
    File = open(LogName, "r")
except IOError, Msg:
    sys.stderr.write("%s: Cannot open file %s for reading: %s" % (Cmd,LogName,Msg))
    sys.exit(10)
Lines = File.readlines();
File.close()
sys.stderr.write("%s: %d lines read.\n" % (Cmd,len(Lines)))

# parse the contents

Examples = [ ]
Dict = { }
for l in Lines:
    (Host,u,Usr,s,Sys,e,Ela,c,Cpu,x,Exi,CmdLine) = [t(s) for t,s in zip((str,float,str,float,str,float,str,str,str,int,str,str),l.split())]
#   print Host,u,s,e,c,x,CmdLine
    Host = Host[:-1]	# get rid of ':' in host name
    c = int(c[:-1])	# get rid of '%' in CPU utilization
    if CmdLine.startswith("./"):
	CmdLine = CmdLine[2:]
#   print Host,u,s,e,c,x,CmdLine
    if CmdLine in Examples:
	sys.stderr.write("%s: Multiple results for '%s'\n" % (Cmd,CmdLine))
    else:
	Examples.append(CmdLine)
	Dict[CmdLine] = (Host,u,s,e,c,x)
sys.stderr.write("%s: %d result records.\n" % (Cmd,len(Dict)))

# execute the checks

NumChecks = 0
NumTests = 0
NumWarnings = 0
for check in Checks:
    (Name,Tests) = check
    Example = Name+"_seq"
    if Example in Examples:
	Data = Dict[Example]
	seq = Perf(Data[0],Data[1],Data[2],Data[3],Data[4],Data[5])
    else:
	sys.stderr.write("%s: Skipping '%s', no data available!\n"
				% (Cmd,Example))
	continue
    Example = Name+"_ooo"
    if Example in Examples:
	Data = Dict[Example]
	ooo = Perf(Data[0],Data[1],Data[2],Data[3],Data[4],Data[5])
    else:
	sys.stderr.write("%s: Skipping '%s', no data available!\n"
				% (Cmd,Example))
	continue
#   print seq, ooo
    NumChecks += 1
    for Test in Tests:
	NumTests += 1
	if not eval(Test):
	    NumWarnings += 1
	    print "\nWARNING: for %s, test '%s' failed!" % (Name,Test)
	    PrintLogLine(seq, Name+"_seq")
	    PrintLogLine(ooo, Name+"_ooo")

# result

if NumWarnings:
    print "\nWARNING: %d tests failed!" % NumWarnings
sys.stderr.write("%s: %d examples checked, a total of %d tests performed.\n"
		% (Cmd, NumChecks, NumTests))
sys.stderr.write("%s: Done.\n" % Cmd)

sys.exit(NumWarnings)

# EOF
