// list.cpp: Spencer Kam (SK), Spring 2019
//
// modifications: (most recent first)
// 07/27/19 RD  reduce excessive printing
// 07/26/19 RD  integrated from 'devel_sk' branch

#include <iostream>
#include <unordered_set>

#include "risc.h"

std::unordered_set<char> mod_commands;
std::unordered_set<char> hie_commands;
std::unordered_set<char> var_commands;
std::unordered_set<char> int_commands;
std::unordered_set<char> prim_commands;
int default_mode = 0;
std::string path;

void myPrintDesign(risc::Design* d);

void myPrintDesign2(risc::Design* d);
void myPrintModule(risc::ModuleVector* vec, std::unordered_set<char>* commands);

int main(int argc, char* argv[])
{
  if(argc < 2){
    printf("Incorrect number of arguments\n");
    return -1;
  }
  
  std::string helpCheck(argv[1]);
  if(helpCheck == "-h"){
    std::cout << "\nUsage: " << argv[0] << " <command>-<subcommands> <SystemC model>\n\n";

    std::cout << "Command::  m: Modules\n";
    std::cout << "  Subcommands:\n";
    std::cout << "  -m: minimal mode\n";
    std::cout << "  -t: include module types\n";
    std::cout << "  -v: print global variables\n";
    std::cout << "  -s: print submodules\n";
    std::cout << "  -p: print ports\n";
    std::cout << "  -h: print hierarchical channels\n";
    std::cout << "  -f: print member functions\n";
    std::cout << "  -l: print source location of the module\n\n";

    std::cout << "Command::  h: Hierarchical Channels\n";
    std::cout << "  Subcommands:\n";
    std::cout << "  -m: minimal mode\n";
    std::cout << "  -t: include module types\n";
    std::cout << "  -v: print global variables\n";
    std::cout << "  -s: print submodules\n";
    std::cout << "  -p: print ports\n";
    std::cout << "  -h: print hierarchical channels\n";
    std::cout << "  -f: print member functions\n";
    std::cout << "  -l: print source location of the channel\n\n";

    std::cout << "Command::  v: Global variables\n";
    std::cout << "  Subcommands:\n";
    std::cout << "  -m: minimal mode\n";
    std::cout << "  -t: include types\n";
    std::cout << "  -l: print source location of the variable\n\n";

    std::cout << "Command::  p: Primitive Channels\n";
    std::cout << "  Subcommands:\n";
    std::cout << "  -m: minimal mode\n";
    std::cout << "  -t: include types\n";
    std::cout << "  -l: print source location of the channel\n\n";

    std::cout << "Command::  i: Interfaces\n";
    std::cout << "  Subcommands:\n";
    std::cout << "  -m: minimal mode\n";
    std::cout << "  -a: print out all interfaces (also from headers)\n";
    std::cout << "  -l: print source location of the interface\n\n";
    return 0;
  }

//std::cout << std::endl
//          << "start processing file: " __FILE__
//          << " ..."
//          << std::endl;
  std::cout << argv[0] << ": Starting..." << std::endl;

  char* actual_arg[3];
  path = "-I";
  path += getenv(SYSTEMC_LW_HOME_ENV);
  path += "/include/";
  //printf("%s", path.c_str());
  
  actual_arg[0] = argv[0];
  actual_arg[1] = (char*)path.c_str();
  actual_arg[2] = argv[argc-1];

  risc::Design d(3, actual_arg);

  //path is changed for use later  
  path.erase(0, 2); 

  //create hash map to store the commands
  for(int i = 1; i < argc-1; i++){
    std::string temp_str(argv[i]);
    /*if(temp_str.length()){
      printf("Valid commands must be in the form of 'command'-'specifier'\n");
      return -1;
    }*/
    char temp_char = temp_str.at(1);
    if(temp_char != '-'){
      printf("Valid commands must be in the form of 'command'-'subcommand'\n");
      return -1;
    }
    temp_char = temp_str.at(0);
    switch(temp_char){
      case 'm':
        for(int j = 1; j < temp_str.length(); j++){
          mod_commands.insert(temp_str.at(j));
        }
        default_mode++;
        break;
      case 'h':
        for(int j = 1; j < temp_str.length(); j++){
          hie_commands.insert(temp_str.at(j));
        }
        default_mode++;
        break;
      case 'v':
        for(int j = 1; j < temp_str.length(); j++){
          var_commands.insert(temp_str.at(j));
        }
        default_mode++;
        break;
      case 'i':
        for(int j = 1; j < temp_str.length(); j++){
          int_commands.insert(temp_str.at(j));
        }
        default_mode++;
        break;
      case 'p':
        for(int j = 1; j < temp_str.length(); j++){
          prim_commands.insert(temp_str.at(j));
        }
        default_mode++;
        break;
      default:
        std::cout << "Command: '" << temp_char << "' was not recognized\n";
    }
  }

  //myPrintDesign(&d);
  myPrintDesign2(&d);

//std::cout << std::endl;

//std::cout << "end processing file: " __FILE__
//          << " ..."
//          << std::endl
//          << std::endl;
  std::cout << argv[0] << ": Done." << std::endl;

  return 0;
}

//This version of myPrint reads command lines
void myPrintDesign2(risc::Design* d){
//printf("======================================  Printing Design Now  ===================================\n\n");
  

  //Modules
  if(!mod_commands.empty() || default_mode == 0){
    printf("=== Modules ===\n\n");
    
    risc::ModuleVector* vectorPointer = (risc::ModuleVector*)&(d->module_definitions_);
    std::unordered_set<char>* commands = &(mod_commands);
    
    myPrintModule(vectorPointer, commands);
    
  }

  //hierarchical channels
  if(!hie_commands.empty() || default_mode == 0){   
    printf("=== Hierarchical Channels ===\n\n");

    risc::ModuleVector* vectorPointer = (risc::ModuleVector*)&(d->hierarchical_channel_);
    std::unordered_set<char>* commands = &(hie_commands);
    
    myPrintModule(vectorPointer, commands);
  } 

  //global variables
  if(!var_commands.empty() || default_mode == 0){
    printf("=== Global Variables ===\n\n");
    std::string temp;
    int minimal = (var_commands.find('m') != var_commands.end())? 1 : 0;
    
    for(int i = 0; i < d->global_variables_.size(); i++){
      printf("  (%2d) ", i);
      //include types
      if(var_commands.find('t') != var_commands.end() || var_commands.size()-minimal == 1 || default_mode == 0){
        //minimal?
        if(minimal == 1){
          printf("%s ", d->global_variables_[i]->get_ast_type_name().c_str()); 
        }else{
          printf("%-30s ", d->global_variables_[i]->get_ast_type_name().c_str());
        }
      }
      temp = d->global_variables_[i]->get_name();
      //minimal?
      if(minimal == 1){
        printf("%s ", temp.c_str());
      }else{
        printf("%-22s", temp.c_str());
      }

      //Include file and line location
      if(var_commands.find('l') != var_commands.end() || var_commands.size()-minimal == 1 || default_mode == 0){
        //minimal?
        if(minimal == 1){
          printf("line: %d %s", d->global_variables_[i]->get_line_number(), d->global_variables_[i]->get_file_name().c_str());
        }
        else{
          printf("line: %-6d %s", d->global_variables_[i]->get_line_number(), d->global_variables_[i]->get_file_name().c_str());
        }
      }
      printf("\n");
    }
    printf("\n\n");
  }

  //primitive channels
  if(!prim_commands.empty() || default_mode == 0){
    printf("=== Primitive Channels ===\n\n");
    int minimal = (prim_commands.find('m') != prim_commands.end())? 1 : 0;
    int print = 1;
    int count = 0;

    for(int i = 0; i < d->primitive_channel_.size(); i++){
    
      if(print == 1){
        printf("(%2d) ", count);
        count++;
        //If user wants types included
        if(prim_commands.find('t') != prim_commands.end() || prim_commands.size()-minimal == 1 || default_mode == 0){
          std::string temp;
          temp = d->primitive_channel_[i]->get_ast_type_name();
  
          //Minimal form?
          if(minimal == 1){
            printf("%s ", temp.c_str());
          }else{
            printf("%-30s ", temp.c_str());
          }
        }
  
        //Print name.  Minimal?
        
        if(minimal == 1){
          printf("%s ", d->primitive_channel_[i]->get_name().c_str());
        }else{
          printf("%-22s ", d->primitive_channel_[i]->get_name().c_str());
        }
          
        //include file location and lines
        if(prim_commands.find('l') != prim_commands.end() || prim_commands.size()-minimal == 1 || default_mode == 0){
          //minimal?
          if(minimal == 1){
            printf("line: %d ", d->primitive_channel_[i]->get_line_number());
          }else{
            printf("line: %-6d ", d->primitive_channel_[i]->get_line_number());
          }
          printf("%s", d->primitive_channel_[i]->get_file_name().c_str());
        }
        printf("\n");
      }
    }
    printf("\n\n");
  }

  //interfaces
  if(!int_commands.empty() || default_mode == 0){
    printf("=== Interfaces ===\n\n");
    int minimal = (int_commands.find('m') != int_commands.end())? 1 : 0;
    int print = 1;
    int count = 0;
    
    for(int i = 0; i < d->interfaces_.size(); i++){
      //get the file location
      std::string tempF = d->interfaces_[i]->get_file_name();
      //If the user specifies subcommand 'a', print will always be set
      if(int_commands.find('a') != int_commands.end()){
        print = 1;
      }else{
      //Otherwise, print is only set if file location was not included from a header file.
        if(tempF.find(path) != std::string::npos || tempF == "compilerGenerated"){
          print = 0;
        }
        else{
          print = 1;
        }
      }
      
      if(print == 1){
        //minimal?
        if(minimal == 1){
          printf("(%2d) %s ", count, d->interfaces_[i]->get_name().c_str());
        }else{
          printf("(%2d) %-30s ", count, d->interfaces_[i]->get_name().c_str());
        }
        count++;
        
        //include file and line location
        if(int_commands.find('l') != int_commands.end() || int_commands.size()-minimal == 1 || default_mode == 0){
          //minimal?
          if(minimal == 1){
            printf("line: %d ", d->interfaces_[i]->get_line_number());
          }else{
            printf("line: %-6d ", d->interfaces_[i]->get_line_number());
          }
          printf("%s", d->interfaces_[i]->get_file_name().c_str());
        }
        printf("\n");
      }
    }
    printf("\n\n");
  }
  
}

void myPrintModule(risc::ModuleVector* vec, std::unordered_set<char>* commands){
  int minimal = (commands->find('m') != commands->end())? 1 : 0;

  for(int i = 0; i < vec->size() ; i++){
      //always print index of current module
      printf("(%2.d) ", i+1);

      //'t' means print type
      if(commands->find('t') != commands->end() || commands->size()-minimal == 1 || default_mode == 0){
        //minimal?
        if(minimal == 1){
          printf("%s ", (*vec)[i]->get_ast_type_name().c_str());
        }else{
          printf("%-25s ", (*vec)[i]->get_ast_type_name().c_str());
        }
      }

      //Always print out name of current module
      //minimal?
      if(minimal == 1){
        printf("%s ", (*vec)[i]->get_name().c_str());
      }else{
        printf("%-15s ", (*vec)[i]->get_name().c_str());
      }
      
      //include file and line location
      if(commands->find('l') != commands->end() || commands->size()-minimal == 1 || default_mode == 0){
        //minimal?
        if(minimal == 1){
          printf("line: %d ", (*vec)[i]->get_line_number());
          }else{
            printf("line: %-6d ", (*vec)[i]->get_line_number());
          }
          printf("%s", (*vec)[i]->get_file_name().c_str());
        }
        printf("\n");


      //'v' means print variables
      if(commands->find('v') != commands->end() || commands->size()-minimal == 1 || default_mode == 0){

        if((*vec)[i]->variables_.size() != 0) {
          //minimal?
          if(minimal == 1){
            printf("       Variables:\n");
          }else{
            printf("\n    ***Variables:______________________________________________________________\n");
            printf("         %15s %-24s | (Name)\n", " ", "(Type)");
          }

          for(int j = 0;
                  j < (*vec)[i]->variables_.size();
                  j++) {
              //minimal?
              if(minimal == 1){
                printf("         %s %s\n", (*vec)[i]->variables_[j]->get_ast_type_name().c_str() ,
                                                 (*vec)[i]->variables_[j]->get_name().c_str());
              }else{
                printf("         %-40s | %-15s\n", (*vec)[i]->variables_[j]->get_ast_type_name().c_str() , 
                                                 (*vec)[i]->variables_[j]->get_name().c_str());
              }
            }
          if(minimal == 0)
            printf("       End of Variables________________________________________________________\n");
        }
      }
      //'s' means print submodules
      if(commands->find('s') != commands->end() || commands->size()-minimal == 1 || default_mode == 0){

        if((*vec)[i]->sub_modules_.size() != 0) {
          //minimal?
          if(minimal == 1){
            printf("       Submodules:\n");
          }else{
            printf("\n    ***Submodules:_____________________________________________________________\n ");
            printf("         %15s %-24s| (Name)\n", " ", "(Type)");
          }

          for(int j = 0;
                  j < (*vec)[i]->sub_modules_.size();
                  j++) {
              //minimal?
              if(minimal == 1){
                printf("         %s %s\n", (*vec)[i]->sub_modules_[j]->get_ast_type_name().c_str() ,
                                                  (*vec)[i]->sub_modules_[j]->get_name().c_str());
              }else{
                printf("         %-40s | %-15s\n", (*vec)[i]->sub_modules_[j]->get_ast_type_name().c_str() ,
                                                  (*vec)[i]->sub_modules_[j]->get_name().c_str());
              }
            }
          if(minimal == 0)
            printf("       End of Submodules_______________________________________________________\n");
        }
      }
      //'p' means print ports
      if(commands->find('p') != commands->end() || commands->size()-minimal == 1 || default_mode == 0){
        
        if( ((*vec)[i]->inports_.size() != 0) ||
            ((*vec)[i]->outports_.size() != 0) ||
            ((*vec)[i]->inoutports_.size() != 0)) {

          //minimal?
          if(minimal == 1){
            printf("       Ports:\n");
          }else{
            printf("\n    ***Ports:__________________________________________________________________\n");
          }
          
          //input ports
          if(!(*vec)[i]->inports_.empty()){
            //minimal?
            if(minimal == 1){
              printf("         In:\n");
            }else{
              printf("\n         In=====\n");
            }
            for(int
                iter  = 0;
                iter < (*vec)[i]->inports_.size();
                iter++) {
        
              //minimal?
              if(minimal == 1){
                printf("         %2d) %s %s\n", iter, (*vec)[i]->inports_[iter]->get_ast_type_name().c_str(),
                                                     (*vec)[i]->inports_[iter]->get_name().c_str());
              }else{
                printf("         %-2d) %-30s %-30s\n", iter, (*vec)[i]->inports_[iter]->get_ast_type_name().c_str(),
                                                     (*vec)[i]->inports_[iter]->get_name().c_str());
              }
        
              printf("           Mapped Channel:\n");
              std::set<risc::Object*> mapped_channels;
              (*vec)[i]->inports_[iter]->get_mapped_channels(mapped_channels);
              std::list<std::string> sorted_names;
              for(std::set<risc::Object*>::iterator
                  channel_iter  = mapped_channels.begin();
                  channel_iter != mapped_channels.end();
                  channel_iter++){
                    sorted_names.push_back((*channel_iter)->get_name());   
              }
              sorted_names.sort();
              for(std::list<std::string>::iterator it = sorted_names.begin(); it != sorted_names.end(); it++) {
                    printf("               %s\n", (*it).c_str());;
              }
              
              printf("           Mapped Channel Types:\n");
              std::set<risc::Class*> mapped_channel_types;
              (*vec)[i]->inports_[iter]->get_mapped_channel_types(mapped_channel_types);
              std::list<std::string> sorted_names2;
              for(std::set<risc::Class*>::iterator
                  channel_iter  = mapped_channel_types.begin();
                  channel_iter != mapped_channel_types.end();
                  channel_iter++){
                    sorted_names2.push_back((*channel_iter)->get_name());   
              }
              sorted_names2.sort();
              for(std::list<std::string>::iterator it = sorted_names2.begin(); it != sorted_names2.end(); it++) {
                    printf("               %s\n", (*it).c_str());;
              }
          
            }
          }
          //output ports
          if(!(*vec)[i]->outports_.empty()){
            //minimal?
            if(minimal == 1){
              printf("         Out:\n");
            }else{
              printf("\n         Out====\n");
            }
            for(int
                iter  = 0;
                iter < (*vec)[i]->outports_.size();
                iter++) {
              //minimal?
              if(minimal == 1){
                printf("         %d) %s %s\n", iter, (*vec)[i]->outports_[iter]->get_ast_type_name().c_str(),
                                                     (*vec)[i]->outports_[iter]->get_name().c_str());
              }else{
                printf("         %-2d) %-30s %-30s\n", iter, (*vec)[i]->outports_[iter]->get_ast_type_name().c_str(),
                                                     (*vec)[i]->outports_[iter]->get_name().c_str());
              }
        
              printf("           Mapped Channel:\n");
              std::set<risc::Object*> mapped_channels;
              (*vec)[i]->outports_[iter]->get_mapped_channels(mapped_channels);
              std::list<std::string> sorted_names;
              for(std::set<risc::Object*>::iterator
                  channel_iter  = mapped_channels.begin();
                  channel_iter != mapped_channels.end();
                  channel_iter++){
                    sorted_names.push_back((*channel_iter)->get_name());   
              }
              sorted_names.sort();
              for(std::list<std::string>::iterator it = sorted_names.begin(); it != sorted_names.end(); it++) {
                    printf("               %s\n", (*it).c_str());;
              }
              
              printf("           Mapped Channel Types:\n");
              std::set<risc::Class*> mapped_channel_types;
              (*vec)[i]->outports_[iter]->get_mapped_channel_types(mapped_channel_types);
              std::list<std::string> sorted_names2;
              for(std::set<risc::Class*>::iterator
                  channel_iter  = mapped_channel_types.begin();
                  channel_iter != mapped_channel_types.end();
                  channel_iter++){
                    sorted_names2.push_back((*channel_iter)->get_name());   
              }
              sorted_names2.sort();
              for(std::list<std::string>::iterator it = sorted_names2.begin(); it != sorted_names2.end(); it++) {
                    printf("               %s\n", (*it).c_str());;
              }
          
            }
          }
          //Inout ports
          if(!(*vec)[i]->inoutports_.empty()){
            //minimal?
            if(minimal == 1){
              printf("         inOut:\n");
            }else{
              printf("\n         inOut==\n");
            }
            for(int
                iter  = 0;
                iter < (*vec)[i]->inoutports_.size();
                iter++) {
              //minimal?
              if(minimal == 1){
                printf("         %d) %s %s\n", iter, (*vec)[i]->inoutports_[iter]->get_ast_type_name().c_str(),
                                                     (*vec)[i]->inoutports_[iter]->get_name().c_str());
              }else{
                printf("         %-2d) %-30s %-30s\n", iter, (*vec)[i]->inoutports_[iter]->get_ast_type_name().c_str(),
                                                     (*vec)[i]->inoutports_[iter]->get_name().c_str());
              }
        
              printf("           Mapped Channel:\n");
              std::set<risc::Object*> mapped_channels;
              (*vec)[i]->inoutports_[iter]->get_mapped_channels(mapped_channels);
              std::list<std::string> sorted_names;
              for(std::set<risc::Object*>::iterator
                  channel_iter  = mapped_channels.begin();
                  channel_iter != mapped_channels.end();
                  channel_iter++){
                    sorted_names.push_back((*channel_iter)->get_name());   
              }
              sorted_names.sort();
              for(std::list<std::string>::iterator it = sorted_names.begin(); it != sorted_names.end(); it++) {
                    printf("               %s\n", (*it).c_str());;
              }
              
              printf("           Mapped Channel Types:\n");
              std::set<risc::Class*> mapped_channel_types;
              (*vec)[i]->inoutports_[iter]->get_mapped_channel_types(mapped_channel_types);
              std::list<std::string> sorted_names2;
              for(std::set<risc::Class*>::iterator
                  channel_iter  = mapped_channel_types.begin();
                  channel_iter != mapped_channel_types.end();
                  channel_iter++){
                    sorted_names2.push_back((*channel_iter)->get_name());   
              }
              sorted_names2.sort();
              for(std::list<std::string>::iterator it = sorted_names2.begin(); it != sorted_names2.end(); it++) {
                    printf("               %s\n", (*it).c_str());;
              }
          
            }
          }
        if(minimal == 0)
          printf("       End of Ports____________________________________________________________\n"); 
        }
      }
      //module hierarchical Channels
      if(commands->find('h') != commands->end() || commands->size()-minimal == 1 || default_mode == 0){
        
        if(!(*vec)[i]->sub_hier_channel_.empty()){
          //minimal?
          if(minimal == 1){
            printf("       Hierarchical Channels:\n");
          }else{
            printf("\n    ***Hierarchical Channels___________________________________________________\n");
            printf("         %15s %-24s | (Name)\n", " ", "(Type)");
          }

          for(int j = 0;
                  j < (*vec)[i]->sub_hier_channel_.size();
                  j++){
            //minimal?
            if(minimal == 1){
              printf("         %s %s\n", (*vec)[i]->sub_hier_channel_[j]->get_ast_type_name().c_str(),
                                             (*vec)[i]->sub_hier_channel_[j]->get_name().c_str());
            }else{
              printf("         %-40s | %-s\n", (*vec)[i]->sub_hier_channel_[j]->get_ast_type_name().c_str(), 
                                             (*vec)[i]->sub_hier_channel_[j]->get_name().c_str());
            }
          }
        if(minimal == 0)
        printf("       End of Hierarchical Channels____________________________________________\n");
        }
      }
      //module member functions
      if(commands->find('f') != commands->end() || commands->size()-minimal == 1 || default_mode == 0){

        if(!(*vec)[i]->functions_.empty()){
          //minimal?
          if(minimal == 1){
            printf("       Member Functions:\n");
          }else{
            printf("\n    ***Member Functions________________________________________________________\n");
            printf("         %15s %-24s | (Name)\n", " ", "(Return Type)");
          }

          for(int j = 0;
                  j < (*vec)[i]->functions_.size();
                  j++){
            //minimal?
            if(minimal == 1){
              printf("         %s %s\n", (*vec)[i]->functions_[j]->get_ast_type_name().c_str(),
                                             (*vec)[i]->functions_[j]->get_name().c_str());
            }else{
              printf("         %-40s | %-s\n", (*vec)[i]->functions_[j]->get_ast_type_name().c_str(),
                                             (*vec)[i]->functions_[j]->get_name().c_str());
            }
          }
        if(minimal == 0)
        printf("       End of Member Functions_________________________________________________\n");
        }
      }
      if(minimal == 0)
      printf("\n\n");
    }
}

/*void myPrintDesign(risc::Design* d){
  //Print out single variables first.
  printf("=======================================  MY PRINT IS STARTING HERE ===========================\n\n");
  printf("---------------------------------------      Single Elements       -------------------------------\n");
  printf("Dynamic Analysis File    = %s\n", d->dynamic_analysis_filename_.c_str());
  printf("Instrumentor Output File = %s\n", d->instrumentor_output_filename_.c_str());
  printf("PSG Mode                 = %d\n", d->psg_mode);
  printf("Psg Count                = %d\n", d->psg_count);

  printf("Path Instance Mapper:_______________________________\n");
  //printf("ID:  %d", d->path_instance_mapper_->get_id());
  printf("The Actual Mapping: \n");
  d->path_instance_mapper_->print_mapping();

  printf("\n\n");
  printf("Mapping Info:_______________________________________\n");
  //d->mapping_info_.print();
  printf("Still Being worked on\n\n");

  //Print the contents of each vecotr next  
  printf("\n\n-----------------------------------         Vectors            -------------------------------\n");


  printf("________________PSG Segment Count___________________\n");
  for(int i = 0; i < d->psg_seg_count.size(); i++){
    printf("  (%d) Count: %d\n", d->psg_seg_count[i]);
  }
  printf("\n\n");

  printf("________________PSG File Names_______________________\n");
  for(int i = 0; i < d->psg_file_names_.size(); i++){
    printf("  (%d) %s\n", i, d->psg_file_names_[i].c_str());
  }
  printf("\n\n");


  printf("________________Module Definitions:_______________ \n");
  for(int i = 0; i < d->module_definitions_.size() ; i++){
    printf("(%d) %s    -------------------------------\n", i, d->module_definitions_[i]->get_name().c_str());
    d->module_definitions_[i]->print_module_definition();
  }
  printf("\n\n");
  
  printf("________________Top Modules:_______________________\n");
  for(int i = 0; i < d->top_modules_.size() ; i++){
    printf("  Instance name (%d): %s \n", i, d->top_modules_[i]->get_ast_type_name().c_str());
  }
  printf("\n\n");

  printf("________________Primitive Channels:_________________\n");
  for(int i = 0; i < d->primitive_channel_.size() ; i++){
    printf("  Primitve channel (%d): %s \n", i , d->primitive_channel_[i]->get_name().c_str());
  }
  printf("\n\n");

  printf("________________Hierarchical Channels:______________\n");
  for(int i = 0; i < d->hierarchical_channel_.size() ; i++){
  printf("________________Interfaces__________________________\n");
  for(int i = 0; i < d->interfaces_.size() ; i++){
    printf("  Interface (%d): %s \n", i, d->interfaces_[i]->get_name().c_str());
  }
  printf("\n\n");

  std::string* temp2 = NULL;
  printf("________________Variables___________________________\n");
  for(int i = 0; i < d->global_variables_.size()  ; i++){
    temp2 = &(d->global_variables_[i]->get_name());
    printf(" (%d) Name: %s \n", i , (*temp2).c_str());
    printf("      Type: %s \n\n", d->global_variables_[i]->get_ast_type_name().c_str());
  }
  printf("\n\n");

  printf("________________Functions____________________________\n");
  d->print_all_global_functions();
  printf("\n\n");


  printf("________________Input Files__________________________\n");
  for(int i = 0; i < d->input_files_.size(); i++){
    printf("  (%d) %s\n", i, d->input_files_[i].c_str());
  }
  printf("\n\n");


  printf("________________Loaded Threads_______________________\n");
  for(int i = 0; i < d->loaded_threads.size(); i++){
    printf("  (%d) Name: %s----------------------------\n", i, d->loaded_threads[i]->get_name().c_str());
    d->loaded_threads[i]->print_event_list();
  } 
  printf("\n\n");

  
  printf("________________Simulation Threads___________________\n");
  for(int i = 0; i < d->simulation_threads.size(); i++){
    printf("  (%d) Name: %s----------------------------\n", i, d->simulation_threads[i]->get_name().c_str());
    d->simulation_threads[i]->print_event_list();
  }
  printf("\n\n");

  printf("________________Loaded Module Instances_______________\n");
  for(int i = 0; i < d->loaded_module_instances.size(); i++){
    printf("  (%d) Name: %s\n", d->loaded_module_instances[i]->get_name().c_str());
    printf("       Type: %s\n", d->loaded_module_instances[i]->get_ast_type_name().c_str());
  }
  printf("\n\n");


}*/

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
