#include "data_conflict_table.h"

#include <fstream>

#include "mapped_variable.h"
#include "segment.h"
#include "../tools/misc.h"
#include "../tools/variable_helper.h"
#include "../internal_representation/design.h"
#include "../internal_representation/dynamic_internal_representation.h"
#include "../internal_representation/dynamic_path_instance_mapper.h"
#include "../internal_representation/internal_representation_helper.h"
#include "../internal_representation/static_path_instance_mapper.h"
#include "port_call_path.h"
#include "../internal_representation/object.h"

//#define DEBUG   // (RD, 11/30/18)

static int get_id(risc::InstanceTree path, SgVariableSymbol *symbol);

static bool FoundValidConflict(           // bug fix (RD, 11/29/18)
            risc::sg::MappedVariable &temp,
            int inst_id,
            bool debugging,
            std::set<risc::sg::Conflict> &conflicts);

extern DynamicModule *ast_root;

risc::sg::DataConflictTable::DataConflictTable(
    SegmentGraph &graph,
    PathInstanceMapper *path_instance_mapper,
    bool debugging):
  ConflictTable(graph, path_instance_mapper, debugging)
{
  // here we perform a read/write analysis on each segment in the graph
  std::pair<VertexIterator, VertexIterator> vp;
  for(vp = boost::vertices(segment_graph_.graph_);
      vp.first != vp.second; ++vp.first) {

    segment_graph_.graph_[*vp.first].analyze_segment();
  }
  //jun 7
  // for(vp = boost::vertices(segment_graph_.graph_);
  //       vp.first != vp.second; ++vp.first) 
  // {
  //   std::cout << "In segment " << segment_graph_.graph_[*vp.first].id_ << std::endl;
  //   for(std::set<SymbolWithPath>::iterator
  //     read_iter  = segment_graph_.graph_[*vp.first].read_variables_.begin(),
  //     read_end   = segment_graph_.graph_[*vp.first].read_variables_.end();
  //     read_iter != read_end;
  //     read_iter++) 
  //   {
  //     std::cout << "var is " <<SageInterface::get_name( read_iter->symbol_) << ( read_iter->symbol_) << std::endl;
  //     const_cast<risc::sg::PortCallPath*>(&(read_iter->pcp_))->print();
  //     std::cout << "!swp.pcp_.port_.empty() is " << std::boolalpha << !(read_iter->pcp_).port_.empty() << std::endl;
  //   }
  // }


  determine_conflict_table();
}

void
risc::sg::DataConflictTable::determine_conflict_table()
{
  bool dynamic_analysis = path_instance_mapper_->perform_dynamic_analysis();
  for(int row=0; row < size_of_conflict_table_; row++) {

    // Here, we are getting the segment id and instance id with the associated
    // row index 'row'.
    std::pair<int, int> row_id_and_instance =
      index_to_segment_and_instance_id(row);

    int row_seg = row_id_and_instance.first;
    int row_inst = row_id_and_instance.second;

    for(int column=0; column <= row; column++) {

      // Here, we are getting the segment id and instance id with the associated
      // column index 'column'.
      std::pair<int, int> column_id_and_instance =
        index_to_segment_and_instance_id(column);

      int column_seg = column_id_and_instance.first;
      int column_inst = column_id_and_instance.second;
#ifdef DEBUG
      printf("DEBUG: row %d col %d = row:seg,inst %d,%d col:seg,inst %d,%d.\n",
                    row,column, row_seg,row_inst, column_seg,column_inst);
#endif

#if 0 // DONT_HANDLE_DIAGONAL_LINE_IN_DATA_CONFLICT_TABLE_SPECIALLY (RD, 04/12/19)

      if(column_seg == row_seg && row_inst == column_inst) {

        // If only read access in this segment, there will be no conflict
        if(segment_graph_.graph_[segment_id_to_vertex_descriptor(column_seg)].
            write_variables_.empty()) {

          conflict_table_[row][column] = false;
        } else {

          bool has_conflict = false;

          for(std::set<SymbolWithPath>::iterator
              write_iter  = segment_graph_.
                              graph_[segment_id_to_vertex_descriptor(row_seg)].
                                write_variables_.begin();
              write_iter != segment_graph_.
                              graph_[segment_id_to_vertex_descriptor(row_seg)].
                                write_variables_.end();
              write_iter++) {
            
            
            MappedVariableList mapped_variables
              = get_mapped_variable(write_iter->symbol_, row_inst,
                  dynamic_analysis);

            for(MappedVariableList::iterator
                mapped_variable  = mapped_variables.begin();
                mapped_variable != mapped_variables.end();
                mapped_variable++) {

              SgVariableSymbol *variable_symbol = (*mapped_variable).symbol_;

              if(is_conflict_free_variable(variable_symbol)) {
                continue;
              }

              if((tools::is_member_variable(variable_symbol)) ||

                  (tools::is_local_variable(variable_symbol) &&
                  tools::is_static_variable(variable_symbol)) ||

                  tools::is_global_variable(variable_symbol) ||
                  tools::is_namespace_variable(variable_symbol)) {

                if((tools::is_channel_variable((*mapped_variable).symbol_)
                  || tools::is_socket_variable(*write_iter))
                    && !tools::is_reference_variable(write_iter->symbol_)) {

                  mapped_variable->instance_id_
                    = get_channel_id(*write_iter, row_inst, dynamic_analysis);
                }

                has_conflict = true;    // (RD: WW conflict for segment with itself)
                                        // RD: no need to distinguish sc_signal here

                if(debugging_) {

                  // We are displaying the mapped variable instead of the
                  // regular variable ... just in case of a reference
                  Conflict conflict(*mapped_variable);
                  conflict_info_[column][row].conflict_variables_.
                    insert(conflict);
                  assert(row == column); // RD: no need to do this twice!
                //conflict_info_[row][column].conflict_variables_.
                //  insert(conflict);
                }
              }
            }
          }

          conflict_table_[row][column] = has_conflict;
        }

        if(segment_graph_.graph_[segment_id_to_vertex_descriptor(row_seg)]
                .calls_conflict_function_
            || segment_graph_.graph_[segment_id_to_vertex_descriptor(row_seg)]
                .deref_pointer_write_){

          if(debugging_) {
            conflict_info_[row][column].conflict_variables_ =
              segment_graph_.graph_[segment_id_to_vertex_descriptor(row_seg)].
                conflicts_;
          }
        }

        conflict_table_[row][column] |=
          segment_graph_.graph_[segment_id_to_vertex_descriptor(row_seg)]
            .calls_conflict_function_;

        conflict_table_[row][column] |=
          segment_graph_.graph_[segment_id_to_vertex_descriptor(row_seg)]
            .deref_pointer_write_;

        // we don't need a read check in this case because we have
        // the same segment and instance id
        // (RD: diagonal line in conflict table)

#ifdef DEBUG
        printf("DEBUG: data_conflict_between(row==col %d = seg,inst %d,%d is %s)\n",
                    row, row_seg,row_inst, (conflict_table_[row][column] ? "TRUE" : "FALSE"));
#endif
      } else {

        // Here we perform the conflict analysis of the individual parts
        // (RD: left/bottom triangle in conflict table)

#endif // DONT_HANDLE_DIAGONAL_LINE_IN_DATA_CONFLICT_TABLE_SPECIALLY (RD, 04/12/19)

#ifdef DEBUG
        printf("DEBUG: data_conflict_between(row %d col %d = row:seg,inst %d,%d col:seg,inst %d,%d)\n",
                    row,column, row_seg,row_inst, column_seg,column_inst);
#endif
        conflict_table_[row][column] = data_conflict_between(
            segment_graph_.graph_[segment_id_to_vertex_descriptor(row_seg)],
            segment_graph_.graph_[segment_id_to_vertex_descriptor(column_seg)],
            row,
            column,
            row_inst,
            column_inst,
            dynamic_analysis);
        conflict_table_[column][row] =   // (RD: copy to right/top triangle)
            conflict_table_[row][column];

#if 0 // DONT_HANDLE_DIAGONAL_LINE_IN_DATA_CONFLICT_TABLE_SPECIALLY (RD, 04/12/19)
      }
#endif // DONT_HANDLE_DIAGONAL_LINE_IN_DATA_CONFLICT_TABLE_SPECIALLY (RD, 04/12/19)
    }
  }
}

void
risc::sg::DataConflictTable::color_members_of_class(
    const MappedVariable &variable,
    DataConflictAstAttributes::Status color)
{
  // we get the type of the variable
  SgInitializedName *var_decl_stmt = variable.symbol_->get_declaration();
  SgType *type = var_decl_stmt->get_type();
  SgDeclarationStatement *type_decl_stmt = type->getAssociatedDeclaration();

  if(!type_decl_stmt) {
    return;
  }

  SgClassDeclaration *class_decl = isSgClassDeclaration(type_decl_stmt);

  if(!class_decl) {
    return;
  }

  class_decl = isSgClassDeclaration(class_decl->get_definingDeclaration());

  SgClassDefinition *class_def = class_decl->get_definition();

  MappedVariable temp;

  for(std::vector<SgDeclarationStatement*>::iterator
      member_iter  = class_def->get_members().begin();
      member_iter != class_def->get_members().end();
      member_iter++) {

    SgVariableDeclaration *var_decl = isSgVariableDeclaration(*member_iter);
    if(var_decl) {

      if(!var_decl->get_decl_item(0)->get_storageModifier().isStatic()) {

        SgVariableSymbol *var_symbol =
          isSgVariableSymbol(var_decl->get_definition()->get_vardefn()->
                              search_for_symbol_from_symbol_table());

        if(isSgClassType(var_decl->get_decl_item(0)->get_type())) {

          temp.symbol_ = var_symbol;
          temp.instance_id_ = variable.instance_id_;
          color_members_of_class(temp, color);
        }

        DataConflictAstAttributes *data_conflict_attribute =
          dynamic_cast<DataConflictAstAttributes*>(
              var_symbol->getAttribute("data_conflict"));

        if(data_conflict_attribute != NULL) {
          if(color == DataConflictAstAttributes::UNUSED) {
            data_conflict_attribute->coloring_read_.clear();
            data_conflict_attribute->coloring_write_.clear();
          } else {
            if(color == DataConflictAstAttributes::READ) {
              data_conflict_attribute->coloring_read_.insert(variable.instance_id_);
            } else {
              data_conflict_attribute->coloring_write_.insert(variable.instance_id_);
            }
          }
        }
      }
    }
  }
}

bool
risc::sg::DataConflictTable::check_members_of_class(
    const MappedVariable &mapped_variable,
    DataConflictAstAttributes::Status color,
    std::set<Conflict> &conflicts)
{
  SgInitializedName *var_decl_stmt = mapped_variable.symbol_->get_declaration();
  SgType *type = var_decl_stmt->get_type();
  SgDeclarationStatement *type_decl_stmt = type->getAssociatedDeclaration();

  if(!type_decl_stmt) {
    return false;
  }

  // we get the type of the variable
  SgClassDeclaration *class_decl = isSgClassDeclaration(type_decl_stmt);

  if(!class_decl) {
    return false;
  }

  class_decl = isSgClassDeclaration(class_decl->get_definingDeclaration());

  SgClassDefinition *class_def = class_decl->get_definition();

  bool result = false;

  for(std::vector<SgDeclarationStatement*>::iterator
      member_iter  = class_def->get_members().begin();
      member_iter != class_def->get_members().end();
      member_iter++) {

    SgVariableDeclaration *var_decl = isSgVariableDeclaration(*member_iter);

    if(var_decl) {
      if(!var_decl->get_decl_item(0)->get_storageModifier().isStatic()) {

        SgVariableSymbol *var_symbol =
          isSgVariableSymbol(var_decl->get_definition()->get_vardefn()->
                              search_for_symbol_from_symbol_table());

        //color_members_of_class
        if(isSgClassType(var_decl->get_decl_item(0)->get_type())) {

          MappedVariable temp(var_symbol, mapped_variable.instance_id_);
          if(check_members_of_class(temp, color, conflicts))  {

            result = true;
          }
        }

        DataConflictAstAttributes *data_conflict_attribute =
          dynamic_cast<DataConflictAstAttributes*>(
              var_symbol->getAttribute("data_conflict"));

        if(data_conflict_attribute != NULL) {
          if(color == DataConflictAstAttributes::READ) {
            for(std::set<int>::iterator
                iter  = data_conflict_attribute->coloring_read_.begin();
                iter != data_conflict_attribute->coloring_read_.end();
                iter++) {

              if(*iter == mapped_variable.instance_id_) {
                if(debugging_) {
                  conflicts.insert(MappedVariable(var_symbol,
                        mapped_variable.instance_id_));
                }
                result = true;
              }
            }
          }

          if(color == DataConflictAstAttributes::WRITE) {
            for(std::set<int>::iterator
                iter  = data_conflict_attribute->coloring_write_.begin();
                iter != data_conflict_attribute->coloring_write_.end();
                iter++) {

              if(*iter == mapped_variable.instance_id_) {
                if(debugging_) {
                  conflicts.insert(MappedVariable(var_symbol,
                        mapped_variable.instance_id_));
                }
                result = true;
              }
            }
          }
        }
      }
    }
  }
  return result;
}

bool
risc::sg::DataConflictTable::data_conflict_between( // (RD: determine one entry in the
    Segment &first_seg,                             //      left/bottom triangle in conflict table)
    Segment &second_seg,
    int row, int column,
    int inst_id_first, int inst_id_second,
    bool dynamic_analysis)
{
  bool seg1_conflict = false;                 // (RD: first check for general conflicts, e.g.
  bool seg2_conflict = false;                 //      pointer accesses or conflicting functions)

  if(first_seg.contains_sc_stop_)
  {
    seg1_conflict = true;

  }  

  if(   first_seg.contains_sc_stop_
     || first_seg.calls_conflict_function_
     || first_seg.deref_pointer_write_
     || (first_seg.deref_pointer_read_
         && !second_seg.write_variables_.empty())) {
    seg1_conflict = true;
    if(debugging_) {
      conflict_info_[row][column].conflict_variables_.insert(
            first_seg.conflicts_.begin(), first_seg.conflicts_.end());
      conflict_info_[column][row].conflict_variables_.insert(
            first_seg.conflicts_.begin(), first_seg.conflicts_.end());
      conflict_info_[row][column].contains_sc_stop_ |= first_seg.contains_sc_stop_;
      conflict_info_[column][row].contains_sc_stop_ |= first_seg.contains_sc_stop_;
    }
  }

  if(     second_seg.contains_sc_stop_
     ||   second_seg.calls_conflict_function_
     || second_seg.deref_pointer_write_
     || (second_seg.deref_pointer_read_
         && !first_seg.write_variables_.empty())) {
    seg2_conflict = true;
    if(debugging_) {
      conflict_info_[row][column].conflict_variables_.insert(
            second_seg.conflicts_.begin(), second_seg.conflicts_.end());
      conflict_info_[column][row].conflict_variables_.insert(
            second_seg.conflicts_.begin(), second_seg.conflicts_.end());
      conflict_info_[row][column].contains_sc_stop_ |= second_seg.contains_sc_stop_;
      conflict_info_[column][row].contains_sc_stop_ |= second_seg.contains_sc_stop_;
    }
  }

  bool data_conflict = false;                 // (RD: then check for specific data conflicts, e.g.
  std::set<Conflict> conflicts;               //      RW,WR,WW conflicts for shared variables)
  MappedVariable temp;                        // NOTE: even with the new std::set data structures,
                                              //       this color matching is probably more
  risc::Design *d = risc::get_design();       //       efficient than using set::find() or similar
                                              //       (RD, 11/29/18)

  for(std::set<SymbolWithPath>::iterator             // (RD: mark 1st segment READ vars)
      read_iter  = first_seg.read_variables_.begin(),
      read_end   = first_seg.read_variables_.end();
      read_iter != read_end;
      read_iter++) {

    if (tools::is_sc_signal(        // RD: sc_signal forces SYSC_SYNC_PAR_SIM,
            read_iter->symbol_)) {  // RD: cannot have any R conflict,
      continue;                     // RD: so skip it!
    }
    if(!tools::is_reference_variable(read_iter->symbol_)) { // (a) normal variable
        temp.symbol_ = read_iter->symbol_;
        temp.instance_id_ = inst_id_first;

        // std::cout << "normal variable " << SageInterface::get_name(temp.symbol_) << temp.symbol_ << std::endl;
        // std::cout << "!*read_iter.pcp_.port_.empty() is " << std::boolalpha << !(*read_iter).pcp_.port_.empty() << std::endl;
        if(tools::is_stack_variable(temp.symbol_)) {
          continue;
        }
        if(is_conflict_free_variable(temp.symbol_)) {
          continue;
        }
        if(tools::is_channel_variable(temp.symbol_) 
          || tools::is_socket_variable(*read_iter)) {
          temp.instance_id_
            = get_channel_id(*read_iter, temp.instance_id_, dynamic_analysis);
        }
        color_symbol(temp, DataConflictAstAttributes::READ);

    } else {                                                // (b) reference variable
      MappedVariableList mapped_variables
        = get_mapped_variable(read_iter->symbol_, inst_id_first,
            dynamic_analysis);
      for(MappedVariableList::iterator
          mapped_variable  = mapped_variables.begin();
          mapped_variable != mapped_variables.end();
          mapped_variable++) {
        if(tools::is_stack_variable((*mapped_variable).symbol_)) {
          continue;
        }
        if(is_conflict_free_variable((*mapped_variable).symbol_)) {
          continue;
        }
        if(tools::is_channel_variable(read_iter->symbol_)
          || tools::is_socket_variable(*read_iter)) {
          int channel_id
            = get_channel_id(*read_iter, inst_id_first, dynamic_analysis);
          // get all mapped modules to this channel
          std::list<std::pair<int, InstanceTree>> result
            = d->mapping_info_.get_mapped_module(
                channel_id,
                SageInterface::getEnclosingClassDefinition(
                  read_iter->symbol_));
          // iterating over all mapped modules
          for(std::list<std::pair<int, InstanceTree>>::iterator
              iter  = result.begin();
              iter != result.end();
              iter++) {
            int sybmol_instance
              = get_id(iter->second, (*mapped_variable).symbol_);
            if (sybmol_instance != -2) {
              mapped_variable->instance_id_ = sybmol_instance;
              color_symbol(*mapped_variable, DataConflictAstAttributes::READ);
            }
          }
        } else {
          color_symbol(*mapped_variable, DataConflictAstAttributes::READ);
        }
      }
    }
  }

  for(std::set<SymbolWithPath>::iterator             // (RD: mark 1st segment WRITE vars)
      write_iter  = first_seg.write_variables_.begin(),
      write_end   = first_seg.write_variables_.end();
      write_iter != write_end;
      write_iter++) {

    if(!tools::is_reference_variable(write_iter->symbol_)) {  // (a) normal variable
        temp.symbol_ = write_iter->symbol_;
        temp.instance_id_ = inst_id_first;
        if(tools::is_stack_variable(temp.symbol_)) {
          continue;
        }
        if(is_conflict_free_variable(temp.symbol_)) {
          continue;
        }
        if(tools::is_channel_variable(temp.symbol_)
          || tools::is_socket_variable(*write_iter)) {
          temp.instance_id_
            = get_channel_id(*write_iter, inst_id_first, dynamic_analysis);
        }
        color_symbol(temp, DataConflictAstAttributes::WRITE);
    } else {                                                  // (b) reference variable
      MappedVariableList mapped_variables
        = get_mapped_variable(write_iter->symbol_, inst_id_first,
            dynamic_analysis);
      for(MappedVariableList::iterator
          mapped_variable  = mapped_variables.begin();
          mapped_variable != mapped_variables.end();
          mapped_variable++) {
        if(tools::is_stack_variable((*mapped_variable).symbol_)) {
          continue;
        }
        if(is_conflict_free_variable((*mapped_variable).symbol_)) {
          continue;
        }
        if(tools::is_channel_variable(write_iter->symbol_)
          || tools::is_socket_variable(*write_iter)) {
          int channel_id
            = get_channel_id(*write_iter, inst_id_first, dynamic_analysis);
          // get all mapped modules to this channel
          std::list<std::pair<int, InstanceTree>> result
            = d->mapping_info_.get_mapped_module(
                channel_id,
                SageInterface::getEnclosingClassDefinition(
                  write_iter->symbol_));
          // iterating over all mapped modules
          for(std::list<std::pair<int, InstanceTree>>::iterator
              iter  = result.begin();
              iter != result.end();
              iter++) {
            int sybmol_instance
              = get_id(iter->second, (*mapped_variable).symbol_);
            if (sybmol_instance != -2) {
              mapped_variable->instance_id_ = sybmol_instance;
              color_symbol(*mapped_variable, DataConflictAstAttributes::WRITE);
            }
          }
        } else {
          color_symbol(*mapped_variable, DataConflictAstAttributes::WRITE);
        }
      }
    }
  }

  for(std::set<SymbolWithPath>::iterator             // (RD: compare 2nd segment READ vars)
      read_iter  = second_seg.read_variables_.begin(),
      read_end   = second_seg.read_variables_.end();
      read_iter != read_end;
      read_iter++) {

    if (tools::is_sc_signal(        // RD: sc_signal forces SYSC_SYNC_PAR_SIM,
            read_iter->symbol_)) {  // RD: cannot have any R conflict,
      continue;                     // RD: so skip it!
    }
    if(!tools::is_reference_variable(read_iter->symbol_)) { // (a) normal variable
      temp.symbol_ = read_iter->symbol_;
      temp.instance_id_ = inst_id_second;
      if(tools::is_stack_variable(temp.symbol_)) {
        continue;
      }
      if(is_conflict_free_variable(temp.symbol_)) {
        continue;
      }
      if(tools::is_channel_variable(temp.symbol_)
        || tools::is_socket_variable(*read_iter)) {
        temp.instance_id_
          = get_channel_id(*read_iter, inst_id_second, dynamic_analysis);
      }
      DataConflictAstAttributes *data_conflict_attribute =
        dynamic_cast<DataConflictAstAttributes*>(
            (temp.symbol_)->getAttribute("data_conflict"));
      if(data_conflict_attribute != NULL) {
        for(std::set<int>::iterator
            iter  = data_conflict_attribute->coloring_write_.begin();
            iter != data_conflict_attribute->coloring_write_.end();
            iter++) {
          if (FoundValidConflict(temp, *iter, debugging_, conflicts)) { // (RD: RW conflict)
            data_conflict = true;
          }
        }
      }
      data_conflict |= check_members_of_class(temp,
                      DataConflictAstAttributes::WRITE, conflicts);

    } else {                                                // (b) reference variable

      MappedVariableList mapped_variables
        = get_mapped_variable(read_iter->symbol_, inst_id_second,
            dynamic_analysis);
      for(MappedVariableList::iterator
          mapped_variable  = mapped_variables.begin();
          mapped_variable != mapped_variables.end();
          mapped_variable++) {
        if(tools::is_stack_variable((*mapped_variable).symbol_)) {
          continue;
        }
        if(is_conflict_free_variable((*mapped_variable).symbol_)) {
          continue;
        }
        std::list<MappedVariable> variables;
        if(tools::is_channel_variable(read_iter->symbol_)
          || tools::is_socket_variable(*read_iter)) {
          //mapped_variable->instance_id_
          //  = get_channel_id(*read_iter, inst_id_second, dynamic_analysis);
          int channel_id
            = get_channel_id(*read_iter, inst_id_second, dynamic_analysis);
          // get all mapped modules to this channel
          std::list<std::pair<int, InstanceTree>> result
            = d->mapping_info_.get_mapped_module(
                channel_id,
                SageInterface::getEnclosingClassDefinition(
                  read_iter->symbol_));
          // Checking the connected path for symbol instance
          for(std::list<std::pair<int, InstanceTree>>::iterator
              iter  = result.begin();
              iter != result.end();
              iter++) {
            int symbol_id = get_id(iter->second, (*mapped_variable).symbol_);
            if (symbol_id != -2) {
              mapped_variable->instance_id_ = symbol_id;
              variables.push_back(*mapped_variable);
            }
          }
        } else {
          variables.push_back(*mapped_variable);
        }
        for(std::list<MappedVariable>::iterator
            var_iter  = variables.begin();
            var_iter != variables.end();
            var_iter++) {
          DataConflictAstAttributes *data_conflict_attribute =
            dynamic_cast<DataConflictAstAttributes*>(
                (var_iter->symbol_)->getAttribute("data_conflict"));
          if(data_conflict_attribute != NULL) {
            for(std::set<int>::iterator
                iter  = data_conflict_attribute->coloring_write_.begin();
                iter != data_conflict_attribute->coloring_write_.end();
                iter++) {
              if (FoundValidConflict(*var_iter, *iter, debugging_, conflicts)) { // (RD: RW conflict)
                data_conflict = true;
              }
            }
          }
          data_conflict |= check_members_of_class(*var_iter,
                          DataConflictAstAttributes::WRITE, conflicts);
        }
      }
    }
  }

  for(std::set<SymbolWithPath>::iterator             // (RD: compare 2nd segment WRITE vars)
      write_iter  = second_seg.write_variables_.begin(),
      write_end   = second_seg.write_variables_.end();
      write_iter != write_end;
      write_iter++) {

    if(!tools::is_reference_variable(write_iter->symbol_)) {  // (a) normal variable
      temp.symbol_ = write_iter->symbol_;
      temp.instance_id_ = inst_id_second;
      if(tools::is_stack_variable(temp.symbol_)) {
        continue;
      }
      if(is_conflict_free_variable(temp.symbol_)) {
        continue;
      }
      if(tools::is_channel_variable(temp.symbol_)
        || tools::is_socket_variable(*write_iter)) {
        temp.instance_id_
          = get_channel_id(*write_iter, temp.instance_id_, dynamic_analysis);
      }
      DataConflictAstAttributes *data_conflict_attribute =
        dynamic_cast<DataConflictAstAttributes*>(
            (temp.symbol_)->getAttribute("data_conflict"));
      if(data_conflict_attribute != NULL) {
        for(std::set<int>::iterator
            iter  = data_conflict_attribute->coloring_read_.begin();
            iter != data_conflict_attribute->coloring_read_.end();
            iter++) {
          if (FoundValidConflict(temp, *iter, debugging_, conflicts)) { // (RD: WR conflict)
            data_conflict = true;
          }
        }
        for(std::set<int>::iterator
            iter  = data_conflict_attribute->coloring_write_.begin();
            iter != data_conflict_attribute->coloring_write_.end();
            iter++) {
          if (FoundValidConflict(temp, *iter, debugging_, conflicts)) { // (RD: WW conflict)
            data_conflict = true;
          }
        }
      }
      data_conflict |= check_members_of_class(temp,
                      DataConflictAstAttributes::WRITE, conflicts);
      data_conflict |= check_members_of_class(temp,
                      DataConflictAstAttributes::READ, conflicts);

    } else {                                                  // (b) reference variable

      MappedVariableList mapped_variables
        = get_mapped_variable(write_iter->symbol_, inst_id_second,
            dynamic_analysis);
      for(MappedVariableList::iterator
          mapped_variable  = mapped_variables.begin();
          mapped_variable != mapped_variables.end();
          mapped_variable++) {
        if(tools::is_stack_variable((*mapped_variable).symbol_)) {
          continue;
        }
        if(is_conflict_free_variable((*mapped_variable).symbol_)) {
          continue;
        }
        std::list<MappedVariable> variables;
        if(tools::is_channel_variable(write_iter->symbol_)
          || tools::is_socket_variable(*write_iter)) {
          int channel_id
            = get_channel_id(*write_iter, inst_id_second, dynamic_analysis);
          // get all mapped modules to this channel
          std::list<std::pair<int, InstanceTree>> result
            = d->mapping_info_.get_mapped_module(
                channel_id,
                SageInterface::getEnclosingClassDefinition(
                  write_iter->symbol_));
          // Checking the connected path for symbol instance
          for(std::list<std::pair<int, InstanceTree>>::iterator
              iter  = result.begin();
              iter != result.end();
              iter++) {
            int symbol_id = get_id(iter->second, (*mapped_variable).symbol_);
            if (symbol_id != -2) {
              mapped_variable->instance_id_ = symbol_id;
              variables.push_back(*mapped_variable);
            }
          }
        } else {
          variables.push_back(*mapped_variable);
        }
        for(std::list<MappedVariable>::iterator
            var_iter  = variables.begin();
            var_iter != variables.end();
            var_iter++) {
          DataConflictAstAttributes *data_conflict_attribute =
            dynamic_cast<DataConflictAstAttributes*>(
                (var_iter->symbol_)->getAttribute("data_conflict"));
          for(std::set<int>::iterator
              iter  = data_conflict_attribute->coloring_read_.begin();
              iter != data_conflict_attribute->coloring_read_.end();
              iter++) {
            if (FoundValidConflict(*var_iter, *iter, debugging_, conflicts)) { // (RD: WR conflict)
              data_conflict = true;
            }
          }
          for(std::set<int>::iterator
              iter  = data_conflict_attribute->coloring_write_.begin();
              iter != data_conflict_attribute->coloring_write_.end();
              iter++) {
            if (FoundValidConflict(*var_iter, *iter, debugging_, conflicts)) { // (RD: WW conflict)
              data_conflict = true;
            }
          }
          data_conflict |= check_members_of_class(*var_iter,
                          DataConflictAstAttributes::WRITE, conflicts);
          data_conflict |= check_members_of_class(*var_iter,
                          DataConflictAstAttributes::READ, conflicts);
        }
      }
    }
  }

  for(std::set<SymbolWithPath>::iterator              // (RD: cleanup, remove READ markers)
      read_iter  = first_seg.read_variables_.begin(),
      read_end   = first_seg.read_variables_.end();
      read_iter != read_end;
      read_iter++) {

    if (tools::is_sc_signal(        // RD: sc_signal forces SYSC_SYNC_PAR_SIM,
            read_iter->symbol_)) {  // RD: cannot have any R conflict,
      continue;                     // RD: so skip it!
    }
    if(!tools::is_reference_variable(read_iter->symbol_)) { // (a) normal variable
      temp.symbol_ = read_iter->symbol_;
      temp.instance_id_ = MappedVariable::InvalidVariable;
      if(tools::is_stack_variable(temp.symbol_)) {
        continue;
      }
      if(is_conflict_free_variable(temp.symbol_)) {
        continue;
      }
      color_symbol(temp, DataConflictAstAttributes::UNUSED);
    } else {                                                // (b) reference variable
      MappedVariableList mapped_variables
        = get_mapped_variable(read_iter->symbol_, inst_id_first,
            dynamic_analysis);
      for(MappedVariableList::iterator
          mapped_variable  = mapped_variables.begin();
          mapped_variable != mapped_variables.end();
          mapped_variable++) {
        if(tools::is_stack_variable((*mapped_variable).symbol_)) {
          continue;
        }
        if(is_conflict_free_variable((*mapped_variable).symbol_)) {
          continue;
        }
        if(tools::is_channel_variable((*mapped_variable).symbol_)
          || tools::is_socket_variable(*read_iter)) {
          mapped_variable->instance_id_
            = get_channel_id(*read_iter, inst_id_first, dynamic_analysis);
        }
        mapped_variable->instance_id_ = MappedVariable::InvalidVariable;
        color_symbol(*mapped_variable, DataConflictAstAttributes::UNUSED);
      }
    }
  }

  for(std::set<SymbolWithPath>::iterator              // (RD: cleanup, remove WRITE markers)
      write_iter  = first_seg.write_variables_.begin(),
      write_end   = first_seg.write_variables_.end();
      write_iter != write_end;
      write_iter++) {
    if(!tools::is_reference_variable(write_iter->symbol_)) {  // (a) normal variable
      temp.symbol_ = write_iter->symbol_;
      temp.instance_id_ = MappedVariable::InvalidVariable;
      if(tools::is_stack_variable(temp.symbol_)) {
        continue;
      }
      if(is_conflict_free_variable(temp.symbol_)) {
        continue;
      }
      color_symbol(temp, DataConflictAstAttributes::UNUSED);
    } else {                                                  // (b) reference variable
      MappedVariableList mapped_variables
        = get_mapped_variable(write_iter->symbol_, inst_id_first,
            dynamic_analysis);
      for(MappedVariableList::iterator
          mapped_variable  = mapped_variables.begin();
          mapped_variable != mapped_variables.end();
          mapped_variable++) {
        if(tools::is_stack_variable((*mapped_variable).symbol_)) {
          continue;
        }
        if(is_conflict_free_variable((*mapped_variable).symbol_)) {
          continue;
        }
        mapped_variable->instance_id_ = MappedVariable::InvalidVariable;
        color_symbol(*mapped_variable, DataConflictAstAttributes::UNUSED);
      }
    }
  }

  if(debugging_) {
    conflict_info_[row][column].conflict_variables_.insert(
            conflicts.begin(), conflicts.end());
    conflict_info_[column][row].conflict_variables_.insert(
            conflicts.begin(), conflicts.end());
  }

  return seg1_conflict || seg2_conflict || data_conflict;
}

bool
risc::sg::DataConflictTable::is_conflict_free_variable(SgVariableSymbol *symbol)
{
  // caching of results
  static std::unordered_map<SgVariableSymbol*, bool> cached_mapping;
  std::unordered_map<SgVariableSymbol*, bool>::iterator iter
    = cached_mapping.find(symbol);
  if(iter != cached_mapping.end()) {
    return iter->second;
  }

  DataConflictAstAttributes *data_conflict_attribute =
    dynamic_cast<DataConflictAstAttributes*>(
        symbol->getAttribute("data_conflict"));

  bool result = false;
  if(data_conflict_attribute == NULL) {

    // this is the case of a mapped variable
    risc::tools::analyze_conflict_free(symbol);
    data_conflict_attribute
      = dynamic_cast<DataConflictAstAttributes*>(
          symbol->getAttribute("data_conflict"));
  }

  result = data_conflict_attribute->is_conflict_free_;

  // caching of results
  cached_mapping[symbol] = result;
  return result;
}

void
risc::sg::DataConflictTable::color_symbol(
      const MappedVariable &variable,
      DataConflictAstAttributes::Status color)
{
  DataConflictAstAttributes *data_conflict_attribute =
    dynamic_cast<DataConflictAstAttributes*>((variable.symbol_)
        ->getAttribute("data_conflict"));

  if(data_conflict_attribute == NULL) {

    data_conflict_attribute = new DataConflictAstAttributes();
    variable.symbol_->setAttribute("data_conflict", data_conflict_attribute);
  }
  if(color == DataConflictAstAttributes::UNUSED) {
    data_conflict_attribute->coloring_read_.clear();
    data_conflict_attribute->coloring_write_.clear();
  } else {
    if(color == DataConflictAstAttributes::READ) {
      data_conflict_attribute->coloring_read_.insert(variable.instance_id_);
    } else {
      data_conflict_attribute->coloring_write_.insert(variable.instance_id_);
    }
  }

  // In the case that we are reading a non pod data type we have to mark
  // all class/struct members as read
  color_members_of_class(variable, color);
}


int get_id(risc::InstanceTree path, SgVariableSymbol *symbol)
{

  // We reduce the path element by element and check if the symbol is
  // in the related module
  while (!path.path_.empty()) {

    // Getting the current variable and the related class definition
    SgVariableDefinition *vardef = path.get_object()->get_ast_node();
    SgClassType *type = isSgClassType(vardef->get_type());

    SgClassDeclaration *classdecl
      = isSgClassDeclaration(type->get_declaration());
    classdecl = isSgClassDeclaration(classdecl->get_definingDeclaration());
    SgClassDefinition *classdef = classdecl->get_definition();


    // check if the symbol is defined in the scope
    if(classdef->symbol_exists(symbol)) {

      return risc::get_design()->path_instance_mapper_->get_id(path);
    }

    // if not found we reduce the instance path by one
    path.path_.pop_back();
  }

  // last check is the global scope
  if(SageInterface::getGlobalScope(symbol)->symbol_exists(symbol)) {
    return -1;  // RD: Tim's magic instance ID for global variables
  }

  // The magic value -2 says that the variable is not found
  return -2;
}

// check a mapped variable for actual conflict
// and record this accurately in the set of conflicts

bool FoundValidConflict(                  // bug fix (RD, 11/29/18)
            risc::sg::MappedVariable &temp,         // candidate variable
            int inst_id,                            // conflicting instance ID
            bool debugging,                         // collect conflicts
            std::set<risc::sg::Conflict> &conflicts // set of collected conflicts
            )
{
  bool conflict = false;  // (keeping Tim's logic as much as possible)

  if(risc::tools::is_member_variable(temp.symbol_)) {
    if(risc::tools::is_static_variable(temp.symbol_)) {
      conflict = true;      // static member variable
      temp.instance_id_ = -3; // RD: new magic instance ID for static member variables (04/11/19)
    }
    if(temp.instance_id_ == inst_id) {
      conflict = true;      // non-static member variable with matching instance ID
    } else {
      // conflict |= false; // non-static member variable with different instance ID
    }
  } else {
    conflict = true;        // global variable (not member variable, nor local variable)
    temp.instance_id_ = -1; // RD: insert magic instance ID for global variables (04/11/19)
  }

  if(debugging && conflict) {
    conflicts.insert(temp);
#ifdef DEBUG
    printf("DEBUG: data_conflict_between() found conflict: var = %s, inst = %d\n",
                  temp.symbol_->get_declaration()->unparseToString().c_str(),
                  temp.instance_id_);
#endif
  }

  return conflict;
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
