#include "tlm.h"                                    // TLM headers
#include "tlm_utils/simple_initiator_socket.h"      // TLM simple initiator socket
#include "tlm_utils/simple_target_socket.h"
#include <string>
#include "systemc.h"
using namespace sc_core;
using namespace std;


/*------------------initiator---------------------*/
class lt_initiator :  public sc_core::sc_module 
{
public:
    lt_initiator( sc_core::sc_module_name name);
  	void initiator_thread (void);
    SC_HAS_PROCESS(lt_initiator);
   
    tlm_utils::simple_initiator_socket<lt_initiator> initiator_socket;
 	int data_init;
 	bool dmi_ptr_valid; //added for dmi
  	tlm::tlm_dmi dmi_data_init; //added for dmi
}; 

lt_initiator::lt_initiator( sc_module_name name )
: sc_module           (name)                      // initialize module name
, initiator_socket    ("initiator_socket")        // initiator socket
{                
	data_init = 42;
	SC_THREAD(initiator_thread);                  
}

void lt_initiator::initiator_thread(void)   ///< initiator thread
{  
	tlm::tlm_generic_payload trans_init; 
	trans_init.set_data_ptr( reinterpret_cast<unsigned char*>(&data_init));
	trans_init.set_data_length(4);
	trans_init.set_dmi_allowed( false ); //added for dmi
	sc_time delay_init = SC_ZERO_TIME;       

	initiator_socket->b_transport(trans_init, delay_init);
	
	//the following code is added for dmi
	if ( trans_init.is_dmi_allowed() )
    {
    	std::cout << "dmi is allowed" << std::endl;
		dmi_data_init.init();
		dmi_ptr_valid = initiator_socket->get_direct_mem_ptr( trans_init, dmi_data_init );
	    
	    unsigned char *ptr_init = dmi_data_init.get_dmi_ptr();
	    for(int i=0; i<10; ++i)
	    {
			wait(2,SC_NS);
	    	memcpy(&data_init, ptr_init, 4);
	    	cout << "getting data using dmi, data_init = " << data_init << endl;
	    }
	}


} 

/*------------------target---------------------*/
class lt_target : public sc_core::sc_module           	
{
public:	  
 
	lt_target( sc_core::sc_module_name   module_name);

	void custom_b_transport( 
		tlm::tlm_generic_payload  &trans,
		sc_core::sc_time          &delay_time
  	);

  	bool custom_get_direct_mem_ptr(
		tlm::tlm_generic_payload  &trans,
	    tlm::tlm_dmi              &dmi_data
    );	

  	void main();
	SC_HAS_PROCESS(lt_target);
	tlm_utils::simple_target_socket<lt_target>  target_socket;

	int target_data;

};

lt_target::lt_target( sc_core::sc_module_name module_name )
: sc_module (module_name)
, target_socket ("target_socket")
{
	target_data = 42;
	target_socket.register_b_transport(this, &lt_target::custom_b_transport);
	target_socket.register_get_direct_mem_ptr(this, &lt_target::custom_get_direct_mem_ptr); //added for dmi


	SC_THREAD(main); 
}

void                                        
lt_target::main(){
	wait(5,SC_NS);
	std::cout << "change target_data to 10 at time 5 sc_ns in target thread main()" << std::endl;
	target_data = 10;
}

void                                        
lt_target::custom_b_transport( 
	tlm::tlm_generic_payload  &trans,
	sc_core::sc_time          &delay_time          
)
{
	unsigned char* ptr = trans.get_data_ptr();
	int len = trans.get_data_length();
	memcpy(ptr, &target_data , len);

	trans.set_dmi_allowed(true);//added for dmi

}

//added for dmi
bool
lt_target::custom_get_direct_mem_ptr(
	tlm::tlm_generic_payload& trans,
    tlm::tlm_dmi& dmi_data
)
{
	// Permit read and write access
	dmi_data.allow_read_write();
	// Set other details of DMI region
	dmi_data.set_dmi_ptr( reinterpret_cast<unsigned char*>( &target_data ) );
	dmi_data.set_start_address( 0 );
	dmi_data.set_end_address( 4 );
	dmi_data.set_read_latency( SC_ZERO_TIME );
	dmi_data.set_write_latency( SC_ZERO_TIME );

	return true;
}

/*------------------top---------------------*/

class lt_top : public sc_core::sc_module                   
{
public:
		
	lt_top ( sc_core::sc_module_name name); 
	lt_target               m_target_1;          
	lt_initiator            m_initiator_1; 
	lt_target               m_target_2;          
	lt_initiator            m_initiator_2;  
};



lt_top::lt_top( sc_core::sc_module_name name )
: sc_core::sc_module(name)
, m_target_1("m_target_1")
, m_initiator_1("m_initiator_1")
, m_target_2("m_target_2")
, m_initiator_2("m_initiator_2")
{
  std::cout << "in top.cpp" << std::endl;
  m_initiator_1.initiator_socket(m_target_1.target_socket);
  m_initiator_2.initiator_socket(m_target_2.target_socket);
}

int sc_main(int   argc, char  *argv[] )
{
  lt_top top("top");                    // instantiate a exmaple top module

  sc_core::sc_start();                  // start the simulation

  return 0;                             // return okay status
}
