#include "tlm.h"                                    // TLM headers
#include "tlm_utils/simple_initiator_socket.h"      // TLM simple initiator socket
#include "tlm_utils/simple_target_socket.h"
#include <string>
#include "systemc.h"
using namespace sc_core;
using namespace std;

class lt_initiator :  public sc_core::sc_module 
{
public:
    lt_initiator( sc_core::sc_module_name name);
  
  	void initiator_thread (void);                    
  
    SC_HAS_PROCESS(lt_initiator);
   
    tlm_utils::simple_initiator_socket<lt_initiator> initiator_socket;
 
    tlm::tlm_response_status gp_status;
}; 


class lt_target : public sc_core::sc_module           	
{
public:	  
 
	lt_target( sc_core::sc_module_name   module_name);
	void foo();
	int a1, a2, a3;
	void custom_b_transport( 
		tlm::tlm_generic_payload  &gp,
		sc_core::sc_time          &delay_time
  	);

	SC_HAS_PROCESS(lt_target);
	tlm_utils::simple_target_socket<lt_target>  m_memory_socket;

	std::string              m_target_memory;
};

class lt_top : public sc_core::sc_module                   
{
public:
		
	lt_top ( sc_core::sc_module_name name); 
	//SimpleBusLT<1, 1>       m_bus;                ///< simple bus
	lt_target               m_lt_target_1;          ///< blocking with convenienece socket
	lt_initiator            m_initiator_1;          ///< instance 1 initiator
};

lt_initiator::lt_initiator( sc_module_name name )
: sc_module           (name)                      // initialize module name
, initiator_socket    ("initiator_socket")        // initiator socket
{                
	SC_THREAD(initiator_thread);                  
}

void lt_initiator::initiator_thread(void)   ///< initiator thread
{  
	tlm::tlm_generic_payload trans; 

	sc_time delay = SC_ZERO_TIME;       // Create delay objects 

	initiator_socket->b_transport(trans, delay);

} 

///Constructor
lt_target::lt_target( sc_core::sc_module_name module_name )
: sc_module (module_name)
, m_memory_socket ("memory_socket_1")
, m_target_memory ("Hello World")
{
  m_memory_socket.register_b_transport(this, &lt_target::custom_b_transport);
}

void                                        
lt_target::custom_b_transport( 
	tlm::tlm_generic_payload  &gp,
	sc_core::sc_time          &delay_time             // delay time 
)
{
  
  int length = gp.get_data_length();
  foo();
  wait(1,SC_NS);
  std::cout << "custom_b_transport called" << std::endl;
    
}

void                                        
lt_target::foo(){
	a1 = 10;
	wait(1,SC_NS);
	a1 = 5;
	if(a1>0){
		a2 = 5;
	}else{
		a3 = 10;
	}
	wait(2,SC_NS);
	a1 = a2+a3;
}

lt_top::lt_top( sc_core::sc_module_name name )
: sc_core::sc_module(name)
, m_lt_target_1("m_lt_target_1")
, m_initiator_1("m_initiator_1")
{
  std::cout << "in top.cpp" << std::endl;
  m_initiator_1.initiator_socket(m_lt_target_1.m_memory_socket);
}

int sc_main(int   argc, char  *argv[] )
{
  lt_top top("top");                    // instantiate a exmaple top module

  sc_core::sc_start();                  // start the simulation

  return 0;                             // return okay status
}
